/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

// Object map / Object click-area module
// Polygon Hit Test code ( HitTestPoly() ) adapted from code (C) Eric Haines
// appearing in Graphics Gems IV, "Point in Polygon Strategies."
// p. 24-46, code: p. 34-45
#include "saga/saga.h"

#include "saga/gfx.h"
#include "saga/console.h"
#include "saga/font.h"
#include "saga/interface.h"
#include "saga/objectmap.h"
#include "saga/actor.h"
#include "saga/scene.h"
#include "saga/isomap.h"
#ifdef SAGA_DEBUG
#include "saga/render.h"
#endif

#include "backends/platform/android/Constants.h"

namespace Saga {

void HitZone::load(SagaEngine *vm, Common::MemoryReadStreamEndian *readStream,
		int index, int sceneNumber) {
	_index = index;
	_flags = readStream->readByte();
	_clickAreas.resize(readStream->readByte());
	_rightButtonVerb = readStream->readByte();
	readStream->readByte(); // pad
	_nameIndex = readStream->readUint16();
	_scriptNumber = readStream->readUint16();

	for (ClickAreas::iterator i = _clickAreas.begin(); i != _clickAreas.end();
			++i) {
		i->resize(readStream->readUint16LE());

		assert(!i->empty());

		for (ClickArea::iterator j = i->begin(); j != i->end(); ++j) {
			j->x = readStream->readSint16();
			j->y = readStream->readSint16();

			// WORKAROUND: bug #1259608: "ITE: Riff ignores command in Ferret merchant center"
			// Apparently ITE Mac version has bug in game data. Both ObjectMap and ActionMap
			// for exit area are little taller (y = 123) and thus Riff goes to exit
			// when clicked on barrel of nails.
			if (vm->getGameId() == GID_ITE) {
				if (sceneNumber == 18 && index == 0
						&& (i == _clickAreas.begin()) && (j == i->begin())
						&& j->y == 123) {
					j->y = 129;
				}
			}
		}
	}
}

bool HitZone::getSpecialPoint(Point &specialPoint) const {
	for (ClickAreas::const_iterator i = _clickAreas.begin();
			i != _clickAreas.end(); ++i) {
		if (i->size() == 1) {
			specialPoint = (*i)[0];
			return true;
		}
	}
	return false;
}

bool HitZone::hitTest(const Point &testPoint) {
	const Point *points;
	uint pointsCount;

	if (_flags & kHitZoneEnabled) {
		for (ClickAreas::const_iterator i = _clickAreas.begin();
				i != _clickAreas.end(); ++i) {
			pointsCount = i->size();
			if (pointsCount < 2) {
				continue;
			}
			points = &i->front();
			if (pointsCount == 2) {
				// Hit-test a box region
				if ((testPoint.x >= points[0].x) && (testPoint.x <= points[1].x)
						&& (testPoint.y >= points[0].y)
						&& (testPoint.y <= points[1].y)) {
					return true;
				}
			} else {
				// Hit-test a polygon
				if (hitTestPoly(points, pointsCount, testPoint)) {
					return true;
				}
			}
		}
	}
	return false;
}

#ifdef SAGA_DEBUG
void HitZone::draw(SagaEngine *vm, int color) {
	int pointsCount, j;
	Location location;
	HitZone::ClickArea tmpPoints;
	const Point *points;
	Point specialPoint1;
	Point specialPoint2;

	for (ClickAreas::const_iterator i = _clickAreas.begin();
			i != _clickAreas.end(); ++i) {
		pointsCount = i->size();
		points = &i->front();
		if (vm->_scene->getFlags() & kSceneFlagISO) {
			tmpPoints.resize(pointsCount);
			for (j = 0; j < pointsCount; j++) {
				location.u() = points[j].x;
				location.v() = points[j].y;
				location.z = 0;
				vm->_isoMap->tileCoordsToScreenPoint(location, tmpPoints[j]);
			}
			points = &tmpPoints.front();
		}

		if (pointsCount == 2) {
			// 2 points represent a box
			vm->_gfx->drawFrame(points[0], points[1], color);
		} else {
			if (pointsCount > 2) {
				// Otherwise draw a polyline
				// Do a full refresh so that the polyline can be shown
				vm->_render->setFullRefresh(true);
				vm->_gfx->drawPolyLine(points, pointsCount, color);
			}
		}
	}
	if (getSpecialPoint(specialPoint1)) {
		specialPoint2 = specialPoint1;
		specialPoint1.x--;
		specialPoint1.y--;
		specialPoint2.x++;
		specialPoint2.y++;
		vm->_gfx->drawFrame(specialPoint1, specialPoint2, color);
	}
}
#endif

// Loads an object map resource ( objects ( clickareas ( points ) ) )
void ObjectMap::load(const ByteArray &resourceData) {

	if (!_hitZoneList.empty()) {
		error("ObjectMap::load _hitZoneList not empty");
	}

	if (resourceData.empty()) {
		return;
	}

	if (resourceData.size() < 4) {
		error("ObjectMap::load wrong resourceLength");
	}

	ByteArrayReadStreamEndian readS(resourceData, _vm->isBigEndian());

	_hitZoneList.resize(readS.readUint16());

	int idx = 0;
	for (HitZoneArray::iterator i = _hitZoneList.begin();
			i != _hitZoneList.end(); ++i) {
		i->load(_vm, &readS, idx++, _vm->_scene->currentSceneNumber());
	}

	populateInteractionHitAreas();
}

void ObjectMap::populateInteractionHitAreas() {

	LOGD("ObjectMap::populateInteractionHitAreas: scene number %d",
			_vm->_scene->currentSceneNumber());

	const Point *points;
	uint pointsCount;

	int count = 0;
	for (int i = 0; i < _hitZoneList.size(); ++i) {
		const HitZone& hitZone = _hitZoneList[i];
		if (hitZone.getFlags() & kHitZoneEnabled) {

			LOGD(
					"ObjectMap::populateInteractionHitAreas: hitzone id %d name %s",
					hitZone.getHitZoneId(),
					_vm->getObjectName(hitZone.getHitZoneId()));

			if (_vm->checkObjectIgnoreList(hitZone.getHitZoneId(),
					_vm->_scene->currentSceneNumber())) {
				LOGD("ObjectMap::populateInteractionHitAreas: ignoring id %d",
						hitZone.getHitZoneId());
				continue;
			}

			const HitZone::ClickAreas& clickAreas = hitZone.getClickAreas();
			for (HitZone::ClickAreas::const_iterator i = clickAreas.begin();
					i != clickAreas.end(); ++i) {
				pointsCount = i->size();
				LOGD(
						"ObjectMap::populateInteractionHitAreas: click area pointCount %d",
						pointsCount);
				if (pointsCount < 2) {
					continue;
				}
				points = &i->front();
				if (pointsCount == 2) {
					LOGD(
							"ObjectMap::populateInteractionHitAreas: rect %d %d %d %d",
							points[0].x, points[0].y, points[1].x, points[1].y);

					mCurrentHotspots[count].mRect = Rect(points[0].x,
							points[0].y, points[1].x, points[1].y);
				} else {

					// Calculate blocking rect
					int leftMost = GAME_SCREEN_WIDTH, topMost =
					GAME_SCREEN_HEIGHT, rightMost = 0, bottomMost = 0;

					for (int p = 0; p < pointsCount; ++p) {
						const Point& point = points[p];
						if (point.x < leftMost)
							leftMost = point.x;
						if (point.x > rightMost)
							rightMost = point.x;

						if (point.y < topMost)
							topMost = point.y;
						if (point.y > bottomMost)
							bottomMost = point.y;

						LOGD(
								"ObjectMap::populateInteractionHitAreas: poly point %d %d",
								point.x, point.y);

					}

					mCurrentHotspots[count].mRect = Rect(leftMost, topMost,
							rightMost, bottomMost);

				}

				// Check if it's possible to hit this area
				if (canSelectClickArea(mCurrentHotspots[count].mRect,
						hitZone.getIndex())) {
					mCurrentHotspots[count].mHotspotType =
							hitZone.getHitZoneId();

					mCurrentHotspots[count].mDisplayPoint =
							_vm->checkObjectForceDisplayPoint(
									hitZone.getHitZoneId(),
									_vm->_scene->currentSceneNumber());

					++count;
				}
			}
		}
	}

	mCurrentHotspotsCount = count;
}

void ObjectMap::clear() {
	_hitZoneList.clear();

	mCurrentHotspotsCount = 0;
}

#ifdef SAGA_DEBUG
void ObjectMap::draw(const Point& testPoint, int color, int color2) {
	int hitZoneIndex;
	Common::String txtBuf;
	Point pickPoint;
	Point textPoint;
	Location pickLocation;
	pickPoint = testPoint;
	if (_vm->_scene->getFlags() & kSceneFlagISO) {
		assert(_vm->_actor->_protagonist);
		pickPoint.y -= _vm->_actor->_protagonist->_location.z;
		_vm->_isoMap->screenPointToTileCoords(pickPoint, pickLocation);
		pickLocation.toScreenPointUV(pickPoint);
	}

	hitZoneIndex = hitTest(pickPoint);

	for (HitZoneArray::iterator i = _hitZoneList.begin();
			i != _hitZoneList.end(); ++i) {
		i->draw(_vm, (hitZoneIndex == i->getIndex()) ? color2 : color);
	}

	if (hitZoneIndex != -1) {
		txtBuf = Common::String::format("hitZone %d", hitZoneIndex);
		textPoint.x = 2;
		textPoint.y = 2;
		_vm->_font->textDraw(kKnownFontSmall, txtBuf.c_str(), textPoint,
				kITEColorBrightWhite, kITEColorBlack, kFontOutline);
	}
}
#endif

int ObjectMap::hitTest(const Point& testPoint, int forcedId) {

	// Loop through all scene objects
	for (HitZoneArray::iterator i = _hitZoneList.begin();
			i != _hitZoneList.end(); ++i) {

		if (forcedId != ID_NOTHING) {
			if (i->getHitZoneId() == forcedId
					&& (i->getFlags() & kHitZoneEnabled)) {
				//	LOGD("ObjectMap::hitTest: using forcedId %d", forcedId);

				return i->getIndex();
			} else {
				continue;
			}
		} else {
			//	LOGD("ObjectMap::hitTest: no forced id");
		}

		if (i->hitTest(testPoint)) {
			return i->getIndex();
		}
	}

	return -1;
}

uint16 ObjectMap::getInteractionHitAreas(Hotspot* rectArray) {
	uint16 finalCount = 0;
	for (int i = 0; i < mCurrentHotspotsCount; ++i) {

		if (_vm->verifyHotspotIsSelectable(mCurrentHotspots[i])) {
			rectArray[finalCount] = mCurrentHotspots[i];
			++finalCount;
		}
	}

	return finalCount;
}

Hotspot ObjectMap::hotspotHitTest(const Point &testPoint) {

	const Point *points;
	uint pointsCount;

	Hotspot result;
	result.mHotspotType = ID_NOTHING;

	int count = 0;
	for (int i = 0; i < _hitZoneList.size(); ++i) {
		const HitZone& hitZone = _hitZoneList[i];
		if (hitZone.getFlags() & kHitZoneEnabled) {

			if (_vm->checkObjectIgnoreList(hitZone.getHitZoneId(),
					_vm->_scene->currentSceneNumber())) {

				continue;
			}

			const HitZone::ClickAreas& clickAreas = hitZone.getClickAreas();
			for (HitZone::ClickAreas::const_iterator i = clickAreas.begin();
					i != clickAreas.end(); ++i) {
				pointsCount = i->size();
				if (pointsCount < 2) {
					continue;
				}
				points = &i->front();
				if (pointsCount == 2) {

					if ((testPoint.x >= points[0].x)
							&& (testPoint.x <= points[1].x)
							&& (testPoint.y >= points[0].y)
							&& (testPoint.y <= points[1].y)) {

						result.mRect = Rect(points[0].x, points[0].y,
								points[1].x, points[1].y);
						result.mHotspotType = hitZone.getHitZoneId();

						result.mDisplayPoint = _vm->checkObjectForceDisplayPoint(
																		hitZone.getHitZoneId(),
																		_vm->_scene->currentSceneNumber());


						return result;
					}

				} else {

					// Hit-test a polygon
					if (hitTestPoly(points, pointsCount, testPoint)) {

						// Calculate blocking rect
						int leftMost = GAME_SCREEN_WIDTH, topMost =
						GAME_SCREEN_HEIGHT, rightMost = 0, bottomMost = 0;

						for (int p = 0; p < pointsCount; ++p) {
							const Point& point = points[p];
							if (point.x < leftMost)
								leftMost = point.x;
							if (point.x > rightMost)
								rightMost = point.x;

							if (point.y < topMost)
								topMost = point.y;
							if (point.y > bottomMost)
								bottomMost = point.y;

						}

						result.mRect = Rect(leftMost, topMost, rightMost,
								bottomMost);
						result.mHotspotType = hitZone.getHitZoneId();

						result.mDisplayPoint = _vm->checkObjectForceDisplayPoint(
																		hitZone.getHitZoneId(),
																		_vm->_scene->currentSceneNumber());


						return result;
					}
				}
			}
		}
	}



	return result;
}

bool ObjectMap::canSelectClickArea(Common::Rect& rect, int hitZoneIndex) {
	LOGD("ObjectMap::canSelectClickArea: checking rect %s for hitZoneIndex %d",
			rect.debugStr(), hitZoneIndex);
	// Check if it's possible to hit this area
	for (int i = rect.left; i <= rect.right; ++i) {
		for (int j = rect.top; j <= rect.bottom; ++j) {
			if (hitTest(Point(i, j)) == hitZoneIndex) {
				return true;
			}
		}
	}

	LOGD("ObjectMap::canSelectClickArea: can't select this area! ignoring");

	return false;
}

void ObjectMap::cmdInfo() {
//	_vm->_console->debugPrintf("%d zone(s) loaded.\n\n", _hitZoneList.size());
}

} // End of namespace Saga
